@@ -16,7 +16,7 @@
 import abc
 from typing import Callable, Dict, List, Optional, Sequence
 
-from jaxlinop import(
+from jaxlinop import (
     LinearOperator,
     DenseLinearOperator,
     DiagonalLinearOperator,
@@ -24,30 +24,260 @@ from jaxlinop import(
 )
 
 import jax.numpy as jnp
-from chex import dataclass
 from jax import vmap
+import jax
 from jaxtyping import Array, Float
 
 from .config import get_defaults
-from .types import PRNGKeyType
+from chex import PRNGKey as PRNGKeyType
+from jaxutils import PyTree
+
 
 JITTER = get_defaults()["jitter"]
 
+
+class AbstractKernelComputation(PyTree):
+    """Abstract class for kernel computations."""
+
+    def __init__(
+        self,
+        kernel_fn: Callable[
+            [Dict, Float[Array, "1 D"], Float[Array, "1 D"]], Array
+        ] = None,
+    ) -> None:
+        self._kernel_fn = kernel_fn
+
+    @property
+    def kernel_fn(
+        self,
+    ) -> Callable[[Dict, Float[Array, "1 D"], Float[Array, "1 D"]], Array]:
+        return self._kernel_fn
+
+    @kernel_fn.setter
+    def kernel_fn(
+        self,
+        kernel_fn: Callable[[Dict, Float[Array, "1 D"], Float[Array, "1 D"]], Array],
+    ) -> None:
+        self._kernel_fn = kernel_fn
+
+    def gram(
+        self,
+        params: Dict,
+        inputs: Float[Array, "N D"],
+    ) -> LinearOperator:
+
+        """Compute Gram covariance operator of the kernel function.
+
+        Args:
+            kernel (AbstractKernel): The kernel function to be evaluated.
+            params (Dict): The parameters of the kernel function.
+            inputs (Float[Array, "N N"]): The inputs to the kernel function.
+
+        Returns:
+            LinearOperator: Gram covariance operator of the kernel function.
+        """
+
+        matrix = self.cross_covariance(params, inputs, inputs)
+
+        return DenseLinearOperator(matrix=matrix)
+
+    @abc.abstractmethod
+    def cross_covariance(
+        self,
+        params: Dict,
+        x: Float[Array, "N D"],
+        y: Float[Array, "M D"],
+    ) -> Float[Array, "N M"]:
+        """For a given kernel, compute the NxM gram matrix on an a pair
+        of input matrices with shape NxD and MxD.
+
+        Args:
+            kernel (AbstractKernel): The kernel for which the cross-covariance
+                matrix should be computed for.
+            params (Dict): The kernel's parameter set.
+            x (Float[Array,"N D"]): The first input matrix.
+            y (Float[Array,"M D"]): The second input matrix.
+
+        Returns:
+            Float[Array, "N M"]: The computed square Gram matrix.
+        """
+        raise NotImplementedError
+
+    def diagonal(
+        self,
+        params: Dict,
+        inputs: Float[Array, "N D"],
+    ) -> DiagonalLinearOperator:
+        """For a given kernel, compute the elementwise diagonal of the
+        NxN gram matrix on an input matrix of shape NxD.
+
+        Args:
+            kernel (AbstractKernel): The kernel for which the variance
+                vector should be computed for.
+            params (Dict): The kernel's parameter set.
+            inputs (Float[Array, "N D"]): The input matrix.
+
+        Returns:
+            LinearOperator: The computed diagonal variance entries.
+        """
+        diag = vmap(lambda x: self._kernel_fn(params, x, x))(inputs)
+
+        return DiagonalLinearOperator(diag=diag)
+
+
+class DenseKernelComputation(AbstractKernelComputation):
+    """Dense kernel computation class. Operations with the kernel assume
+    a dense gram matrix structure.
+    """
+
+    def __init__(
+        self,
+        kernel_fn: Callable[
+            [Dict, Float[Array, "1 D"], Float[Array, "1 D"]], Array
+        ] = None,
+    ) -> None:
+        super().__init__(kernel_fn)
+
+    def cross_covariance(
+        self, params: Dict, x: Float[Array, "N D"], y: Float[Array, "M D"]
+    ) -> Float[Array, "N M"]:
+        """For a given kernel, compute the NxM covariance matrix on a pair of input
+        matrices of shape NxD and MxD.
+
+        Args:
+            kernel (AbstractKernel): The kernel for which the Gram
+                matrix should be computed for.
+            params (Dict): The kernel's parameter set.
+            x (Float[Array,"N D"]): The input matrix.
+            y (Float[Array,"M D"]): The input matrix.
+
+        Returns:
+            CovarianceOperator: The computed square Gram matrix.
+        """
+        cross_cov = vmap(lambda x: vmap(lambda y: self.kernel_fn(params, x, y))(y))(x)
+        return cross_cov
+
+
+class DiagonalKernelComputation(AbstractKernelComputation):
+    def __init__(
+        self,
+        kernel_fn: Callable[
+            [Dict, Float[Array, "1 D"], Float[Array, "1 D"]], Array
+        ] = None,
+    ) -> None:
+        super().__init__(kernel_fn)
+
+    def gram(
+        self,
+        params: Dict,
+        inputs: Float[Array, "N D"],
+    ) -> DiagonalLinearOperator:
+        """For a kernel with diagonal structure, compute the NxN gram matrix on
+        an input matrix of shape NxD.
+
+        Args:
+            kernel (AbstractKernel): The kernel for which the Gram matrix
+                should be computed for.
+            params (Dict): The kernel's parameter set.
+            inputs (Float[Array, "N D"]): The input matrix.
+
+        Returns:
+            CovarianceOperator: The computed square Gram matrix.
+        """
+
+        diag = vmap(lambda x: self.kernel_fn(params, x, x))(inputs)
+
+        return DiagonalLinearOperator(diag=diag)
+
+    def cross_covariance(
+        self, params: Dict, x: Float[Array, "N D"], y: Float[Array, "M D"]
+    ) -> Float[Array, "N M"]:
+        raise ValueError("Cross covariance not defined for diagonal kernels.")
+
+
+class ConstantDiagonalKernelComputation(AbstractKernelComputation):
+    def __init__(
+        self,
+        kernel_fn: Callable[
+            [Dict, Float[Array, "1 D"], Float[Array, "1 D"]], Array
+        ] = None,
+    ) -> None:
+        super().__init__(kernel_fn)
+
+    def gram(
+        self,
+        params: Dict,
+        inputs: Float[Array, "N D"],
+    ) -> ConstantDiagonalLinearOperator:
+        """For a kernel with diagonal structure, compute the NxN gram matrix on
+        an input matrix of shape NxD.
+
+        Args:
+            kernel (AbstractKernel): The kernel for which the Gram matrix
+                should be computed for.
+            params (Dict): The kernel's parameter set.
+            inputs (Float[Array, "N D"]): The input matrix.
+
+        Returns:
+            CovarianceOperator: The computed square Gram matrix.
+        """
+
+        value = self.kernel_fn(params, inputs[0], inputs[0])
+
+        return ConstantDiagonalLinearOperator(value=value, size=inputs.shape[0])
+
+    def diagonal(
+        self,
+        params: Dict,
+        inputs: Float[Array, "N D"],
+    ) -> DiagonalLinearOperator:
+        """For a given kernel, compute the elementwise diagonal of the
+        NxN gram matrix on an input matrix of shape NxD.
+
+        Args:
+            kernel (AbstractKernel): The kernel for which the variance
+                vector should be computed for.
+            params (Dict): The kernel's parameter set.
+            inputs (Float[Array, "N D"]): The input matrix.
+
+        Returns:
+            LinearOperator: The computed diagonal variance entries.
+        """
+
+        diag = vmap(lambda x: self.kernel_fn(params, x, x))(inputs)
+
+        return DiagonalLinearOperator(diag=diag)
+
+    def cross_covariance(
+        self, params: Dict, x: Float[Array, "N D"], y: Float[Array, "M D"]
+    ) -> Float[Array, "N M"]:
+        raise ValueError("Cross covariance not defined for constant diagonal kernels.")
+
+
 ##########################################
 # Abtract classes
 ##########################################
-@dataclass(repr=False)
-class AbstractKernel:
+class AbstractKernel(PyTree):
     """
     Base kernel class"""
 
-    active_dims: Optional[List[int]] = None
-    stationary: Optional[bool] = False
-    spectral: Optional[bool] = False
-    name: Optional[str] = "AbstractKernel"
-
-    def __post_init__(self) -> None:
+    def __init__(
+        self,
+        compute_engine: AbstractKernelComputation = DenseKernelComputation,
+        active_dims: Optional[List[int]] = None,
+        stationary: Optional[bool] = False,
+        spectral: Optional[bool] = False,
+        name: Optional[str] = "AbstractKernel",
+    ) -> None:
+        self.compute_engine = compute_engine
+        self.active_dims = active_dims
+        self.stationary = stationary
+        self.spectral = spectral
+        self.name = name
         self.ndims = 1 if not self.active_dims else len(self.active_dims)
+        compute_engine = self.compute_engine(kernel_fn=self.__call__)
+        self.gram = compute_engine.gram
+        self.cross_covariance = compute_engine.cross_covariance
 
     @abc.abstractmethod
     def __call__(
@@ -123,206 +353,27 @@ class AbstractKernel:
         raise NotImplementedError
 
 
-@dataclass
-class AbstractKernelComputation:
-    """Abstract class for kernel computations."""
-
-    @staticmethod
-    @abc.abstractmethod
-    def gram(
-        kernel: AbstractKernel,
-        params: Dict,
-        inputs: Float[Array, "N D"],
-    ) -> LinearOperator:
-
-        """Compute Gram covariance operator of the kernel function.
-
-        Args:
-            kernel (AbstractKernel): The kernel function to be evaluated.
-            params (Dict): The parameters of the kernel function.
-            inputs (Float[Array, "N N"]): The inputs to the kernel function.
-
-        Returns:
-            LinearOperator: Gram covariance operator of the kernel function.
-        """
-
-        raise NotImplementedError
-
-    @staticmethod
-    def cross_covariance(
-        kernel: AbstractKernel,
-        params: Dict,
-        x: Float[Array, "N D"],
-        y: Float[Array, "M D"],
-    ) -> Float[Array, "N M"]:
-        """For a given kernel, compute the NxM gram matrix on an a pair
-        of input matrices with shape NxD and MxD.
-
-        Args:
-            kernel (AbstractKernel): The kernel for which the cross-covariance
-                matrix should be computed for.
-            params (Dict): The kernel's parameter set.
-            x (Float[Array,"N D"]): The first input matrix.
-            y (Float[Array,"M D"]): The second input matrix.
-
-        Returns:
-            Float[Array, "N M"]: The computed square Gram matrix.
-        """
-
-        cross_cov = vmap(lambda x: vmap(lambda y: kernel(params, x, y))(y))(x)
-
-        return cross_cov
-
-    @staticmethod
-    def diagonal(
-        kernel: AbstractKernel,
-        params: Dict,
-        inputs: Float[Array, "N D"],
-    ) -> DiagonalLinearOperator:
-        """For a given kernel, compute the elementwise diagonal of the
-        NxN gram matrix on an input matrix of shape NxD.
-
-        Args:
-            kernel (AbstractKernel): The kernel for which the variance
-                vector should be computed for.
-            params (Dict): The kernel's parameter set.
-            inputs (Float[Array, "N D"]): The input matrix.
-
-        Returns:
-            LinearOperator: The computed diagonal variance entries.
-        """
-
-        diag = vmap(lambda x: kernel(params, x, x))(inputs)
-
-        return DiagonalLinearOperator(diag=diag)
-
-
-class DenseKernelComputation(AbstractKernelComputation):
-    """Dense kernel computation class. Operations with the kernel assume
-    a dense gram matrix structure.
-    """
-
-    @staticmethod
-    def gram(
-        kernel: AbstractKernel,
-        params: Dict,
-        inputs: Float[Array, "N D"],
-    ) -> DenseLinearOperator:
-        """For a given kernel, compute the NxN gram matrix on an input
-        matrix of shape NxD.
-
-        Args:
-            kernel (AbstractKernel): The kernel for which the Gram
-                matrix should be computed for.
-            params (Dict): The kernel's parameter set.
-            inputs (Float[Array,"N D"]): The input matrix.
-
-        Returns:
-            CovarianceOperator: The computed square Gram matrix.
-        """
-
-        matrix = vmap(lambda x: vmap(lambda y: kernel(params, x, y))(inputs))(inputs)
-
-        return DenseLinearOperator(matrix=matrix)
-
-
-class DiagonalKernelComputation(AbstractKernelComputation):
-    @staticmethod
-    def gram(
-        kernel: AbstractKernel,
-        params: Dict,
-        inputs: Float[Array, "N D"],
-    ) -> DiagonalLinearOperator:
-        """For a kernel with diagonal structure, compute the NxN gram matrix on
-        an input matrix of shape NxD.
-
-        Args:
-            kernel (AbstractKernel): The kernel for which the Gram matrix
-                should be computed for.
-            params (Dict): The kernel's parameter set.
-            inputs (Float[Array, "N D"]): The input matrix.
-
-        Returns:
-            CovarianceOperator: The computed square Gram matrix.
-        """
-
-        diag = vmap(lambda x: kernel(params, x, x))(inputs)
-
-        return DiagonalLinearOperator(diag=diag)
-
-
-class ConstantDiagonalKernelComputation(AbstractKernelComputation):
-    @staticmethod
-    def gram(
-        kernel: AbstractKernel,
-        params: Dict,
-        inputs: Float[Array, "N D"],
-    ) -> ConstantDiagonalLinearOperator:
-        """For a kernel with diagonal structure, compute the NxN gram matrix on
-        an input matrix of shape NxD.
-
-        Args:
-            kernel (AbstractKernel): The kernel for which the Gram matrix
-                should be computed for.
-            params (Dict): The kernel's parameter set.
-            inputs (Float[Array, "N D"]): The input matrix.
-
-        Returns:
-            CovarianceOperator: The computed square Gram matrix.
-        """
-
-        value = kernel(params, inputs[0], inputs[0])
-
-        return ConstantDiagonalLinearOperator(value=value, size=inputs.shape[0])
-
-
-    @staticmethod
-    def diagonal(
-        kernel: AbstractKernel,
-        params: Dict,
-        inputs: Float[Array, "N D"],
-    ) -> DiagonalLinearOperator:
-        """For a given kernel, compute the elementwise diagonal of the
-        NxN gram matrix on an input matrix of shape NxD.
-
-        Args:
-            kernel (AbstractKernel): The kernel for which the variance
-                vector should be computed for.
-            params (Dict): The kernel's parameter set.
-            inputs (Float[Array, "N D"]): The input matrix.
-
-        Returns:
-            LinearOperator: The computed diagonal variance entries.
-        """
-
-        diag = vmap(lambda x: kernel(params, x, x))(inputs)
-
-        return DiagonalLinearOperator(diag=diag)
-
-
-@dataclass
-class _KernelSet:
-    """A mixin class for storing a list of kernels. Useful for combination kernels."""
-
-    kernel_set: List[AbstractKernel]
-
-
-@dataclass
-class CombinationKernel(AbstractKernel, _KernelSet, DenseKernelComputation):
+class CombinationKernel(AbstractKernel):
     """A base class for products or sums of kernels."""
 
-    name: Optional[str] = "Combination kernel"
-    combination_fn: Optional[Callable] = None
+    def __init__(
+        self,
+        kernel_set: List[AbstractKernel],
+        compute_engine: AbstractKernelComputation = DenseKernelComputation,
+        active_dims: Optional[List[int]] = None,
+        stationary: Optional[bool] = False,
+        spectral: Optional[bool] = False,
+        name: Optional[str] = "AbstractKernel",
+    ) -> None:
+        super().__init__(compute_engine, active_dims, stationary, spectral, name)
+        self.kernel_set = kernel_set
+        name: Optional[str] = "Combination kernel"
+        self.combination_fn: Optional[Callable] = None
 
-    def __post_init__(self) -> None:
-        """Set the kernel set to the list of kernels passed to the constructor."""
-        kernels = self.kernel_set
-
-        if not all(isinstance(k, AbstractKernel) for k in kernels):
+        if not all(isinstance(k, AbstractKernel) for k in self.kernel_set):
             raise TypeError("can only combine Kernel instances")  # pragma: no cover
 
-        self.kernel_set: List[AbstractKernel] = []
-        self._set_kernels(kernels)
+        self._set_kernels(self.kernel_set)
 
     def _set_kernels(self, kernels: Sequence[AbstractKernel]) -> None:
         """Combine multiple kernels. Based on GPFlow's Combination kernel."""
@@ -361,33 +412,57 @@ class CombinationKernel(AbstractKernel, _KernelSet, DenseKernelComputation):
         )
 
 
-@dataclass
 class SumKernel(CombinationKernel):
     """A kernel that is the sum of a set of kernels."""
 
-    name: Optional[str] = "Sum kernel"
-    combination_fn: Optional[Callable] = jnp.sum
+    def __init__(
+        self,
+        kernel_set: List[AbstractKernel],
+        compute_engine: AbstractKernelComputation = DenseKernelComputation,
+        active_dims: Optional[List[int]] = None,
+        stationary: Optional[bool] = False,
+        spectral: Optional[bool] = False,
+        name: Optional[str] = "Sum kernel",
+    ) -> None:
+        super().__init__(
+            kernel_set, compute_engine, active_dims, stationary, spectral, name
+        )
+        self.combination_fn: Optional[Callable] = jnp.sum
 
 
-@dataclass
 class ProductKernel(CombinationKernel):
     """A kernel that is the product of a set of kernels."""
 
-    name: Optional[str] = "Product kernel"
-    combination_fn: Optional[Callable] = jnp.prod
+    def __init__(
+        self,
+        kernel_set: List[AbstractKernel],
+        compute_engine: AbstractKernelComputation = DenseKernelComputation,
+        active_dims: Optional[List[int]] = None,
+        stationary: Optional[bool] = False,
+        spectral: Optional[bool] = False,
+        name: Optional[str] = "Product kernel",
+    ) -> None:
+        super().__init__(
+            kernel_set, compute_engine, active_dims, stationary, spectral, name
+        )
+        self.combination_fn: Optional[Callable] = jnp.prod
 
 
 ##########################################
 # Euclidean kernels
 ##########################################
-@dataclass(repr=False)
-class RBF(AbstractKernel, DenseKernelComputation):
+class RBF(AbstractKernel):
     """The Radial Basis Function (RBF) kernel."""
 
-    name: Optional[str] = "Radial basis function kernel"
-
-    def __post_init__(self) -> None:
-        self.ndims = 1 if not self.active_dims else len(self.active_dims)
+    def __init__(
+        self,
+        compute_engine: AbstractKernelComputation = DenseKernelComputation,
+        active_dims: Optional[List[int]] = None,
+        stationary: Optional[bool] = False,
+        spectral: Optional[bool] = False,
+        name: Optional[str] = "Radial basis function kernel",
+    ) -> None:
+        super().__init__(compute_engine, active_dims, stationary, spectral, name)
 
     def __call__(
         self, params: Dict, x: Float[Array, "1 D"], y: Float[Array, "1 D"]
@@ -412,20 +487,25 @@ class RBF(AbstractKernel, DenseKernelComputation):
         return K.squeeze()
 
     def _initialise_params(self, key: PRNGKeyType) -> Dict:
-        return {
+        params = {
             "lengthscale": jnp.array([1.0] * self.ndims),
             "variance": jnp.array([1.0]),
         }
+        return jax.tree_util.tree_map(lambda x: jnp.atleast_1d(x), params)
 
 
-@dataclass(repr=False)
-class Matern12(AbstractKernel, DenseKernelComputation):
+class Matern12(AbstractKernel):
     """The Matérn kernel with smoothness parameter fixed at 0.5."""
 
-    name: Optional[str] = "Matern 1/2"
-
-    def __post_init__(self) -> None:
-        self.ndims = 1 if not self.active_dims else len(self.active_dims)
+    def __init__(
+        self,
+        compute_engine: AbstractKernelComputation = DenseKernelComputation,
+        active_dims: Optional[List[int]] = None,
+        stationary: Optional[bool] = False,
+        spectral: Optional[bool] = False,
+        name: Optional[str] = "Matérn 1/2 kernel",
+    ) -> None:
+        super().__init__(compute_engine, active_dims, stationary, spectral, name)
 
     def __call__(
         self,
@@ -437,7 +517,7 @@ class Matern12(AbstractKernel, DenseKernelComputation):
         lengthscale parameter :math:`\ell` and variance :math:`\sigma^2`
 
         .. math::
-            k(x, y) = \\sigma^2 \\exp \\Bigg( -\\frac{\\lvert x-y \\rvert}{\\ell}  \\Bigg)
+            k(x, y) = \\sigma^2 \\exp \\Bigg( -\\frac{\\lvert x-y \\rvert}{2\\ell^2}  \\Bigg)
 
         Args:
             params (Dict): Parameter set for which the kernel should be evaluated on.
@@ -458,14 +538,18 @@ class Matern12(AbstractKernel, DenseKernelComputation):
         }
 
 
-@dataclass(repr=False)
-class Matern32(AbstractKernel, DenseKernelComputation):
+class Matern32(AbstractKernel):
     """The Matérn kernel with smoothness parameter fixed at 1.5."""
 
-    name: Optional[str] = "Matern 3/2"
-
-    def __post_init__(self) -> None:
-        self.ndims = 1 if not self.active_dims else len(self.active_dims)
+    def __init__(
+        self,
+        compute_engine: AbstractKernelComputation = DenseKernelComputation,
+        active_dims: Optional[List[int]] = None,
+        stationary: Optional[bool] = False,
+        spectral: Optional[bool] = False,
+        name: Optional[str] = "Matern 3/2",
+    ) -> None:
+        super().__init__(compute_engine, active_dims, stationary, spectral, name)
 
     def __call__(
         self,
@@ -477,7 +561,7 @@ class Matern32(AbstractKernel, DenseKernelComputation):
         lengthscale parameter :math:`\ell` and variance :math:`\sigma^2`
 
         .. math::
-            k(x, y) = \\sigma^2 \\exp \\Bigg(1+ \\frac{\\sqrt{3}\\lvert x-y \\rvert}{\\ell}  \\Bigg)\\exp\\Bigg(-\\frac{\\sqrt{3}\\lvert x-y\\rvert}{\\ell} \\Bigg)
+            k(x, y) = \\sigma^2 \\exp \\Bigg(1+ \\frac{\\sqrt{3}\\lvert x-y \\rvert}{\\ell^2}  \\Bigg)\\exp\\Bigg(-\\frac{\\sqrt{3}\\lvert x-y\\rvert}{\\ell^2} \\Bigg)
 
         Args:
             params (Dict): Parameter set for which the kernel should be evaluated on.
@@ -504,14 +588,18 @@ class Matern32(AbstractKernel, DenseKernelComputation):
         }
 
 
-@dataclass(repr=False)
-class Matern52(AbstractKernel, DenseKernelComputation):
+class Matern52(AbstractKernel):
     """The Matérn kernel with smoothness parameter fixed at 2.5."""
 
-    name: Optional[str] = "Matern 5/2"
-
-    def __post_init__(self) -> None:
-        self.ndims = 1 if not self.active_dims else len(self.active_dims)
+    def __init__(
+        self,
+        compute_engine: AbstractKernelComputation = DenseKernelComputation,
+        active_dims: Optional[List[int]] = None,
+        stationary: Optional[bool] = False,
+        spectral: Optional[bool] = False,
+        name: Optional[str] = "Matern 5/2",
+    ) -> None:
+        super().__init__(compute_engine, active_dims, stationary, spectral, name)
 
     def __call__(
         self, params: Dict, x: Float[Array, "1 D"], y: Float[Array, "1 D"]
@@ -520,7 +608,7 @@ class Matern52(AbstractKernel, DenseKernelComputation):
         lengthscale parameter :math:`\ell` and variance :math:`\sigma^2`
 
         .. math::
-            k(x, y) = \\sigma^2 \\exp \\Bigg(1+ \\frac{\\sqrt{5}\\lvert x-y \\rvert}{\\ell} + \\frac{5\\lvert x - y \\rvert^2}{3\\ell^2} \\Bigg)\\exp\\Bigg(-\\frac{\\sqrt{5}\\lvert x-y\\rvert}{\\ell} \\Bigg)
+            k(x, y) = \\sigma^2 \\exp \\Bigg(1+ \\frac{\\sqrt{5}\\lvert x-y \\rvert}{\\ell^2} + \\frac{5\\lvert x - y \\rvert^2}{3\\ell^2} \\Bigg)\\exp\\Bigg(-\\frac{\\sqrt{5}\\lvert x-y\\rvert}{\\ell^2} \\Bigg)
 
         Args:
             params (Dict): Parameter set for which the kernel should be evaluated on.
@@ -547,15 +635,99 @@ class Matern52(AbstractKernel, DenseKernelComputation):
         }
 
 
-@dataclass(repr=False)
-class Polynomial(AbstractKernel, DenseKernelComputation):
+class PoweredExponential(AbstractKernel):
+    """The powered exponential family of kernels.
+
+    Key reference is Diggle and Ribeiro (2007) - "Model-based Geostatistics".
+
+    """
+
+    def __init__(
+        self,
+        compute_engine: AbstractKernelComputation = DenseKernelComputation,
+        active_dims: Optional[List[int]] = None,
+        stationary: Optional[bool] = False,
+        spectral: Optional[bool] = False,
+        name: Optional[str] = "Powered exponential",
+    ) -> None:
+        super().__init__(compute_engine, active_dims, stationary, spectral, name)
+
+    def __call__(self, params: dict, x: jnp.DeviceArray, y: jnp.DeviceArray) -> Array:
+        """Evaluate the kernel on a pair of inputs :math:`(x, y)` with length-scale parameter :math:`\ell`, :math:`\sigma` and power :math:`\kappa`.
+
+        .. math::
+            k(x, y) = \\sigma^2 \\exp \\Bigg( - \\Big( \\frac{\\lVert x - y \\rVert^2}{\\ell^2} \\Big)^\\kappa \\Bigg)
+
+        Args:
+            x (jnp.DeviceArray): The left hand argument of the kernel function's call.
+            y (jnp.DeviceArray): The right hand argument of the kernel function's call
+            params (dict): Parameter set for which the kernel should be evaluated on.
+
+        Returns:
+            Array: The value of :math:`k(x, y)`
+        """
+        x = self.slice_input(x) / params["lengthscale"]
+        y = self.slice_input(y) / params["lengthscale"]
+        K = params["variance"] * jnp.exp(-euclidean_distance(x, y) ** params["power"])
+        return K.squeeze()
+
+    def _initialise_params(self, key: PRNGKeyType) -> Dict:
+        return {
+            "lengthscale": jnp.array([1.0] * self.ndims),
+            "variance": jnp.array([1.0]),
+            "power": jnp.array([1.0]),
+        }
+
+
+class Linear(AbstractKernel):
+    """The linear kernel."""
+
+    def __init__(
+        self,
+        compute_engine: AbstractKernelComputation = DenseKernelComputation,
+        active_dims: Optional[List[int]] = None,
+        stationary: Optional[bool] = False,
+        spectral: Optional[bool] = False,
+        name: Optional[str] = "Linear",
+    ) -> None:
+        super().__init__(compute_engine, active_dims, stationary, spectral, name)
+
+    def __call__(self, params: dict, x: jnp.DeviceArray, y: jnp.DeviceArray) -> Array:
+        """Evaluate the kernel on a pair of inputs :math:`(x, y)` with variance parameter :math:`\sigma`
+
+        .. math::
+            k(x, y) = \\sigma^2 x^{T}y
+
+        Args:
+            x (jnp.DeviceArray): The left hand argument of the kernel function's call.
+            y (jnp.DeviceArray): The right hand argument of the kernel function's call
+            params (dict): Parameter set for which the kernel should be evaluated on.
+        Returns:
+            Array: The value of :math:`k(x, y)`
+        """
+        x = self.slice_input(x)
+        y = self.slice_input(y)
+        K = params["variance"] * jnp.matmul(x.T, y)
+        return K.squeeze()
+
+    def _initialise_params(self, key: PRNGKeyType) -> Dict:
+        return {"variance": jnp.array([1.0])}
+
+
+class Polynomial(AbstractKernel):
     """The Polynomial kernel with variable degree."""
 
-    name: Optional[str] = "Polynomial"
-    degree: int = 1
-
-    def __post_init__(self) -> None:
-        self.ndims = 1 if not self.active_dims else len(self.active_dims)
+    def __init__(
+        self,
+        degree: int = 1,
+        compute_engine: AbstractKernelComputation = DenseKernelComputation,
+        active_dims: Optional[List[int]] = None,
+        stationary: Optional[bool] = False,
+        spectral: Optional[bool] = False,
+        name: Optional[str] = "Polynomial",
+    ) -> None:
+        super().__init__(compute_engine, active_dims, stationary, spectral, name)
+        self.degree = degree
         self.name = f"Polynomial Degree: {self.degree}"
 
     def __call__(
@@ -586,10 +758,9 @@ class Polynomial(AbstractKernel, DenseKernelComputation):
         }
 
 
-@dataclass(repr=False)
 class White(AbstractKernel, ConstantDiagonalKernelComputation):
     def __post_init__(self) -> None:
-        self.ndims = 1 if not self.active_dims else len(self.active_dims)
+        super(White, self).__post_init__()
 
     def __call__(
         self, params: Dict, x: Float[Array, "1 D"], y: Float[Array, "1 D"]
@@ -622,26 +793,167 @@ class White(AbstractKernel, ConstantDiagonalKernelComputation):
         return {"variance": jnp.array([1.0])}
 
 
+class RationalQuadratic(AbstractKernel):
+    def __init__(
+        self,
+        compute_engine: AbstractKernelComputation = DenseKernelComputation,
+        active_dims: Optional[List[int]] = None,
+        stationary: Optional[bool] = False,
+        spectral: Optional[bool] = False,
+        name: Optional[str] = "Rational Quadratic",
+    ) -> None:
+        super().__init__(compute_engine, active_dims, stationary, spectral, name)
+
+    def __call__(self, params: dict, x: jnp.DeviceArray, y: jnp.DeviceArray) -> Array:
+        """Evaluate the kernel on a pair of inputs :math:`(x, y)` with length-scale parameter :math:`\ell` and variance :math:`\sigma`
+
+        .. math::
+            k(x, y) = \\sigma^2 \\exp \\Bigg( 1 + \\frac{\\lVert x - y \\rVert^2_2}{2 \\alpha \\ell^2} \\Bigg)
+
+        Args:
+            x (jnp.DeviceArray): The left hand argument of the kernel function's call.
+            y (jnp.DeviceArray): The right hand argument of the kernel function's call
+            params (dict): Parameter set for which the kernel should be evaluated on.
+        Returns:
+            Array: The value of :math:`k(x, y)`
+        """
+        x = self.slice_input(x) / params["lengthscale"]
+        y = self.slice_input(y) / params["lengthscale"]
+        K = params["variance"] * (
+            1 + 0.5 * squared_distance(x, y) / params["alpha"]
+        ) ** (-params["alpha"])
+        return K.squeeze()
+
+    def _initialise_params(self, key: PRNGKeyType) -> dict:
+        return {
+            "lengthscale": jnp.array([1.0] * self.ndims),
+            "variance": jnp.array([1.0]),
+            "alpha": jnp.array([1.0]),
+        }
+
+
+class Periodic(AbstractKernel):
+    """The periodic kernel.
+
+    Key reference is MacKay 1998 - "Introduction to Gaussian processes".
+    """
+
+    def __init__(
+        self,
+        compute_engine: AbstractKernelComputation = DenseKernelComputation,
+        active_dims: Optional[List[int]] = None,
+        stationary: Optional[bool] = False,
+        spectral: Optional[bool] = False,
+        name: Optional[str] = "Periodic",
+    ) -> None:
+        super().__init__(compute_engine, active_dims, stationary, spectral, name)
+
+    def __call__(self, params: dict, x: jnp.DeviceArray, y: jnp.DeviceArray) -> Array:
+        """Evaluate the kernel on a pair of inputs :math:`(x, y)` with length-scale parameter :math:`\ell` and variance :math:`\sigma`
+
+        .. math::
+            k(x, y) = \\sigma^2 \\exp \\Bigg( -0.5 \\sum_{i=1}^{d} \\Bigg)
+
+        Args:
+            x (jnp.DeviceArray): The left hand argument of the kernel function's call.
+            y (jnp.DeviceArray): The right hand argument of the kernel function's call
+            params (dict): Parameter set for which the kernel should be evaluated on.
+        Returns:
+            Array: The value of :math:`k(x, y)`
+        """
+        x = self.slice_input(x)
+        y = self.slice_input(y)
+        sine_squared = (
+            jnp.sin(jnp.pi * (x - y) / params["period"]) / params["lengthscale"]
+        ) ** 2
+        K = params["variance"] * jnp.exp(-0.5 * jnp.sum(sine_squared, axis=0))
+        return K.squeeze()
+
+    def _initialise_params(self, key: PRNGKeyType) -> Dict:
+        return {
+            "lengthscale": jnp.array([1.0] * self.ndims),
+            "variance": jnp.array([1.0]),
+            "period": jnp.array([1.0] * self.ndims),
+        }
+
+
 ##########################################
 # Graph kernels
 ##########################################
-@dataclass
-class _EigenKernel:
-    laplacian: Float[Array, "N N"]
+class EigenKernelComputation(AbstractKernelComputation):
+    def __init__(
+        self,
+        kernel_fn: Callable[
+            [Dict, Float[Array, "1 D"], Float[Array, "1 D"]], Array
+        ] = None,
+    ) -> None:
+        super().__init__(kernel_fn)
+        self._eigenvalues = None
+        self._eigenvectors = None
+        self._num_verticies = None
+
+    # Define an eigenvalue setter and getter property
+    @property
+    def eigensystem(self) -> Float[Array, "N"]:
+        return self._eigenvalues, self._eigenvectors, self._num_verticies
+
+    @eigensystem.setter
+    def eigensystem(
+        self, eigenvalues: Float[Array, "N"], eigenvectors: Float[Array, "N N"]
+    ) -> None:
+        self._eigenvalues = eigenvalues
+        self._eigenvectors = eigenvectors
+
+    @property
+    def num_vertex(self) -> int:
+        return self._num_verticies
+
+    @num_vertex.setter
+    def num_vertex(self, num_vertex: int) -> None:
+        self._num_verticies = num_vertex
+
+    def _compute_S(self, params):
+        evals, evecs = self.eigensystem
+        S = jnp.power(
+            evals
+            + 2 * params["smoothness"] / params["lengthscale"] / params["lengthscale"],
+            -params["smoothness"],
+        )
+        S = jnp.multiply(S, self.num_vertex / jnp.sum(S))
+        S = jnp.multiply(S, params["variance"])
+        return S
+
+    def cross_covariance(
+        self, params: Dict, x: Float[Array, "N D"], y: Float[Array, "M D"]
+    ) -> Float[Array, "N M"]:
+        S = self._compute_S(params=params)
+        matrix = self.kernel_fn(params, x, y, S=S)
+        return matrix
 
 
-@dataclass
-class GraphKernel(AbstractKernel, _EigenKernel, DenseKernelComputation):
-    name: Optional[str] = "Graph kernel"
-
-    def __post_init__(self) -> None:
-        self.ndims = 1
+class GraphKernel(AbstractKernel):
+    def __init__(
+        self,
+        laplacian: Float[Array, "N N"],
+        compute_engine: EigenKernelComputation = EigenKernelComputation,
+        active_dims: Optional[List[int]] = None,
+        stationary: Optional[bool] = False,
+        spectral: Optional[bool] = False,
+        name: Optional[str] = "Graph kernel",
+    ) -> None:
+        super().__init__(compute_engine, active_dims, stationary, spectral, name)
+        self.laplacian = laplacian
         evals, self.evecs = jnp.linalg.eigh(self.laplacian)
         self.evals = evals.reshape(-1, 1)
-        self.num_vertex = self.laplacian.shape[0]
+        self.compute_engine.eigensystem = self.evals, self.evecs
+        self.compute_engine.num_vertex = self.laplacian.shape[0]
 
     def __call__(
-        self, params: Dict, x: Float[Array, "1 D"], y: Float[Array, "1 D"]
+        self,
+        params: Dict,
+        x: Float[Array, "1 D"],
+        y: Float[Array, "1 D"],
+        **kwargs,
     ) -> Float[Array, "1"]:
         """Evaluate the graph kernel on a pair of vertices :math:`v_i, v_j`.
 
@@ -653,17 +965,11 @@ class GraphKernel(AbstractKernel, _EigenKernel, DenseKernelComputation):
         Returns:
             Float[Array, "1"]: The value of :math:`k(v_i, v_j)`.
         """
-        psi = jnp.power(
-            2 * params["smoothness"] / params["lengthscale"] ** 2 + self.evals,
-            -params["smoothness"],
-        )
-        psi *= self.num_vertex / jnp.sum(psi)
-        x_evec = self.evecs[:, x]
-        y_evec = self.evecs[:, y]
-        kxy = params["variance"] * jnp.sum(
-            jnp.prod(jnp.stack([psi, x_evec, y_evec]).squeeze(), axis=0)
-        )
-        return kxy.squeeze()
+        S = kwargs["S"]
+        Kxx = (jax_gather_nd(self.evecs, x) * S[None, :]) @ jnp.transpose(
+            jax_gather_nd(self.evecs, y)
+        )  # shape (n,n)
+        return Kxx.squeeze()
 
     def _initialise_params(self, key: PRNGKeyType) -> Dict:
         return {
@@ -672,6 +978,10 @@ class GraphKernel(AbstractKernel, _EigenKernel, DenseKernelComputation):
             "smoothness": jnp.array([1.0]),
         }
 
+    @property
+    def num_vertex(self) -> int:
+        return self.compute_engine.num_vertex
+
 
 def squared_distance(
     x: Float[Array, "1 D"], y: Float[Array, "1 D"]
@@ -705,6 +1015,11 @@ def euclidean_distance(
     return jnp.sqrt(jnp.maximum(squared_distance(x, y), 1e-36))
 
 
+def jax_gather_nd(params, indices):
+    tuple_indices = tuple(indices[..., i] for i in range(indices.shape[-1]))
+    return params[tuple_indices]
+
+
 __all__ = [
     "AbstractKernel",
     "CombinationKernel",
@@ -714,6 +1029,9 @@ __all__ = [
     "Matern12",
     "Matern32",
     "Matern52",
+    "Linear",
+    "Periodic",
+    "RationalQuadratic",
     "Polynomial",
     "White",
     "GraphKernel",
